/*
 * Copyright (c) 2010-2024. Institut Pasteur.
 *
 * This file is part of Icy.
 * Icy is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Icy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Icy. If not, see <https://www.gnu.org/licenses/>.
 */

package plugins.adufour.roi;

import org.bioimageanalysis.extension.kernel.roi.roi2d.ROI2DArea;
import org.bioimageanalysis.extension.kernel.roi.roi2d.ROI2DPolygon;
import org.bioimageanalysis.extension.kernel.roi.roi2d.ROI2DShape;
import org.bioimageanalysis.extension.kernel.roi.roi3d.ROI3DArea;
import org.bioimageanalysis.icy.common.geom.point.Point3D;
import org.bioimageanalysis.icy.extension.plugin.annotation_.IcyPluginIcon;
import org.bioimageanalysis.icy.extension.plugin.annotation_.IcyPluginName;
import org.bioimageanalysis.icy.model.roi.ROI;
import org.bioimageanalysis.icy.model.roi.ROI2D;
import org.bioimageanalysis.icy.model.roi.ROI3D;
import org.bioimageanalysis.icy.model.roi.ROIUtil;
import org.bioimageanalysis.icy.model.sequence.Sequence;
import org.jetbrains.annotations.NotNull;
import plugins.adufour.blocks.tools.roi.ROIBlock;
import plugins.adufour.blocks.util.VarList;
import plugins.adufour.ezplug.EzPlug;
import plugins.adufour.ezplug.EzVarBoolean;
import plugins.adufour.ezplug.EzVarSequence;
import plugins.adufour.quickhull.QuickHull2D;
import plugins.adufour.quickhull.QuickHull3D;
import plugins.adufour.roi.mesh.polygon.ROI3DPolygonalMesh;
import plugins.adufour.vars.lang.VarROIArray;

import javax.vecmath.Point3d;
import java.awt.geom.Point2D;
import java.util.Arrays;
import java.util.List;

@IcyPluginName("Convexify")
@IcyPluginIcon(path = "/convexify.png")
public class Convexify extends EzPlug implements ROIBlock {
    private final EzVarSequence input = new EzVarSequence("Input sequence");
    private final EzVarBoolean replace = new EzVarBoolean("Replace existing ROI", false);
    private final VarROIArray roiIN = new VarROIArray("List of ROI");
    private final VarROIArray roiOUT = new VarROIArray("List of ROI");

    @Override
    protected void initialize() {
        addEzComponent(input);
        addEzComponent(replace);
    }

    @Override
    public void clean() {

    }

    @Override
    public void execute() {
        if (!isHeadLess()) {
            roiIN.setValue(new ROI[0]);
            final Sequence s = input.getValue(true);
            roiIN.add(s.getROIs().toArray(new ROI[0]));
        }

        roiOUT.setValue(new ROI[0]);
        for (final ROI roi : roiIN)
            roiOUT.add(createConvexROI(roi));

        if (!isHeadLess()) {
            final Sequence s = input.getValue(true);

            if (replace.getValue())
                s.removeAllROI();

            for (final ROI roi : roiOUT) {
                s.addROI(roi);
            }
        }
    }

    /**
     * Creates a convex ROI from the specified ROI.
     *
     * @param roi the ROI to be converted into a convex version
     * @return a convex version of the ROI. This ROI is currently of type {@link ROI2DPolygon} in
     * 2D, and of type {@link ROI3DArea} in 3D, although the return types may change in the
     * future.
     * @throws IllegalArgumentException if the specified ROI is not supported
     */
    public static @NotNull ROI createConvexROI(final ROI roi) throws IllegalArgumentException {
        ROI output = null;

        try {
            if (roi instanceof final ROI2D roi2d) {
                final List<Point2D> envelope;

                if (roi instanceof ROI2DShape) {
                    envelope = QuickHull2D.computeConvexEnvelope(((ROI2DShape) roi).getPoints());
                }
                else if (roi instanceof ROI2DArea) {
                    final Point2D[] points = ((ROI2DArea) roi).getBooleanMask(true).getContourPoints();
                    envelope = QuickHull2D.computeConvexEnvelope(Arrays.asList(points));
                }
                else {
                    throw new IllegalArgumentException("Cannot compute convex hull for ROI " + roi.getName()
                            + " of type " + roi.getClassName() + ".");
                }

                output = new ROI2DPolygon(envelope);

                // copy position info
                ((ROI2D) output).setT(roi2d.getT());
                ((ROI2D) output).setZ(roi2d.getZ());
                ((ROI2D) output).setC(roi2d.getC());
            }
            else if (roi instanceof final ROI3D roi3d) {
                final Point3D[] points = ((ROI3D) roi).getBooleanMask(true).getContourPoints();

                // convert to vecmath's Point3d
                final Point3d[] pt3d = new Point3d[points.length];
                for (int i = 0; i < points.length; i++) {
                    final Point3D p = points[i];
                    pt3d[i] = new Point3d(p.getX(), p.getY(), p.getZ());
                }

                final QuickHull3D qhull = new QuickHull3D();
                qhull.build(pt3d, pt3d.length);

                // Use a 3D mesh to prepare the final ROI
                output = new ROI3DPolygonalMesh(qhull);

                // copy position info
                ((ROI3D) output).setT(roi3d.getT());
                ((ROI3D) output).setC(roi3d.getC());
            }
        }
        catch (final Throwable t) {
            // can happen with QuickHull3D / ROI3DPolygonalMesh...
            throw new IllegalArgumentException(
                    "Cannot compute convex hull for ROI " + roi.getName() + " of type " + roi.getClassName() + ".", t);
        }

        if (output == null)
            throw new IllegalArgumentException(
                    "Cannot compute convex hull for ROI " + roi.getName() + " of type " + roi.getClassName() + ".");

        ROIUtil.copyROIProperties(roi, output, false);
        output.setName(roi.getName() + " (convex)");

        return output;
    }

    @Override
    public void declareInput(final @NotNull VarList inputMap) {
        inputMap.add("Regions of interest", roiIN);
    }

    @Override
    public void declareOutput(final @NotNull VarList outputMap) {
        outputMap.add("Regions of interest", roiOUT);
    }
}
